/* With non GPL files, use following license */
/*
 * Copyright 2018 Sony Corporation, Sony Video & Sound Products Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/* Otherwise with GPL files, use following license */
/*
 * Copyright 2018 Sony Corporation, Sony Video & Sound Products Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License.
 *
 * THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 * WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 * USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * You should have received a copy of the  GNU General Public License along
 * with this program; if not, write  to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * nandbank.h
 */

#ifndef __NAND_BANK_H__
#define __NAND_BANK_H__

#define NAND_ENTRY_NUM  (32)
#define NAND_ENTRY_LAST_ID (31)
#define NAND_ENTRY_SIZE (64)

#define NAND_TRANSACTION_ID_OFFSET (0)
#define NAND_CHECKSUM_OFFSET (4)

#define NAND_PAGE_SIZE ( 2048 )
#define NAND_PAGE_NUM  ( 64 )

#define NAND_BLOCK_SIZE ( NAND_PAGE_SIZE * NAND_PAGE_NUM )
#define NAND_BLOCK_COUNT ( 1 )

struct ST_PAGE
{
    int id;
    unsigned long addr;
    unsigned int transaction_id;
    unsigned int checksum;

    void (*init)( struct ST_PAGE * _page, unsigned long base_addr, int page_id );
    int  (*load)( struct ST_PAGE * _page, char * p_page_buf );
    int  (*is_match_checksum)( struct ST_PAGE * _page, char * p_page_buf );
    int  (*is_empty)( struct ST_PAGE * _page, char * p_page_buf );

    int (*add_entry)( struct ST_PAGE * _page, char * p_page_buf, const char *key, const void *value, int len, unsigned int transaction_id );
    int (*del_entry)( struct ST_PAGE * _page, char * p_page_buf, const char *key, unsigned int transaction_id );
};


struct ST_BLOCK
{
    int id;
    unsigned long addr;

    int  read_page_id;
    int  write_page_id;
    int  is_page_full;

    struct ST_PAGE page[NAND_PAGE_NUM];

    void (*init)( struct ST_BLOCK * _block, int block_id );
    void (*erase)( struct ST_BLOCK * _block );
    int  (*load_active_page)( struct ST_BLOCK * _block, struct ST_PAGE * p_active_page, char * p_page_buf );
    int  (*add_entry)( struct ST_BLOCK * _block, char * p_page_buf, const char *key, const void *value, int len, unsigned int transaction_id );
    int  (*del_entry)( struct ST_BLOCK * _block, char * p_page_buf, const char *key, unsigned int transaction_id );
};


struct ST_BANK
{
    struct ST_BLOCK block[2];
    int id;
    int active_block_id;
    char page_buf[NAND_ENTRY_NUM][NAND_ENTRY_SIZE];

    void (*init)( struct ST_BANK * _bank, int bank_id );
    
    void (*load_active_page)( struct ST_BANK * _bank );
    void (*dump_active_page)( struct ST_BANK * _bank );

    int (*add_entry)( struct ST_BANK * _bank, const char *key, const void *value, int len);
    int (*get_entry)( struct ST_BANK * _bank, const char *key, void *buf, int maxlen);
    int (*del_entry)( struct ST_BANK * _bank, const char *key );
};

extern void bank_init( struct ST_BANK * _bank, int bank_id );
extern int nandbank_init( void );



#endif//__NAND_BANK_H__
