/*
 * Copyright 2018 Sony Corporation, Sony Video & Sound Products Inc.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * nvreg_drv.c
 */

#include <linux/cdev.h>
#include <asm/uaccess.h>

#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>

#include "DeviceCommon.h"
#include "nvreg_error.h"
#include "nvreg_drv.h"
#include "nandreg.h"

#ifndef __KERNEL__
#define __KERNEL__
#endif

#ifndef MODULE
#define MODULE
#endif



const char * GetIoctlCmdStr( EN_NVREG_IOCTL command )
{
    return 
        command == NVREG_IOCTL__ADD     ? "IOCTL:ADD"   :
        command == NVREG_IOCTL__GET     ? "IOCTL:GET"   :
        command == NVREG_IOCTL__DEL     ? "IOCTL:DEL"   :
        command == NVREG_IOCTL__DUMP    ? "IOCTL:DUMP"  :"????";
}

const char * GetRetStr( int ret )
{
    return 
        ret == NVREG_OK                 ?   "NVREG_OK"                   :
        ret == NVREG_ERR_INTERNAL       ?   "NVREG_ERR_INTERNAL"         :
        ret == NVREG_ERR_INVALID_PARAM  ?   "NVREG_ERR_INVALID_PARAM"    :
        ret == NVREG_ERR_DEV_FILE       ?   "NVREG_ERR_DEV_FILE"         :
        ret == NVREG_ERR_ENTRY_FULL     ?   "NVREG_ERR_ENTRY_FULL"       :
        ret == NVREG_ERR_NO_ENTRY       ?   "NVREG_ERR_NO_ENTRY"         : "???";
}

int nvreg_add( struct file *file, unsigned long arg )
{
    ST_IOCTL_PARAM	env;
    unsigned long ret;
    char key[NVREG_COPY_BUF_SIZE];
    char val[NVREG_COPY_BUF_SIZE];
    int key_len = 0;

    ret = copy_from_user( (void *)&env, (void *)arg, sizeof(ST_IOCTL_PARAM) );
    if( ret < 0 ) printk("[NVREG][add]copy_from_user 1 failed %lu\n", ret ); 

    if( env.key == NULL || env.val == NULL )
    {
        return NVREG_ERR_INVALID_PARAM;
    }
    ret = copy_from_user( key, env.key, NVREG_COPY_BUF_SIZE );
    if( ret < 0 ) printk("[NVREG][add]copy_from_user 2 failed %lu\n", ret ); 
    ret = copy_from_user( val, env.val, NVREG_COPY_BUF_SIZE );
    if( ret < 0 ) printk("[NVREG][add]copy_from_user 3 failed %lu\n", ret ); 

    key_len = strnlen(key, NVREG_COPY_BUF_SIZE);
    if( key_len >= NVREG_COPY_BUF_SIZE )
    {
        printk("[NVREG] ERROR :  the char * key is not terminated with NULL.\n");
        return NVREG_ERR_INVALID_PARAM;
    }

    return nandreg_add( key, val, env.val_len );
}

#ifdef CONFIG_COMPAT
int nvreg_compat_add( struct file *file, unsigned long arg )
{
    ST_IOCTL_PARAM	env;
    ST_IOCTL_PARAM32	env32;
    unsigned long ret;
    char key[NVREG_COPY_BUF_SIZE];
    char val[NVREG_COPY_BUF_SIZE];
    int key_len = 0;

    ret = copy_from_user( (void *)&env32, (void *)arg, sizeof(ST_IOCTL_PARAM32) );
    if( ret < 0 ) printk("[NVREG][add]copy_from_user 1 failed %lu\n", ret ); 

    env.key = compat_ptr(env32.key);
    env.val = compat_ptr(env32.val);
    env.val_len = env32.val_len;

    if( env.key == NULL || env.val == NULL )
    {
        return NVREG_ERR_INVALID_PARAM; 
    }
    
    ret = copy_from_user( key, env.key, NVREG_COPY_BUF_SIZE );
    if( ret < 0 ) printk("[NVREG][add]copy_from_user 2 failed %lu\n", ret ); 
    ret = copy_from_user( val, env.val, NVREG_COPY_BUF_SIZE );
    if( ret < 0 ) printk("[NVREG][add]copy_from_user 3 failed %lu\n", ret ); 

    key_len = strnlen(key, NVREG_COPY_BUF_SIZE);
    if( key_len >= NVREG_COPY_BUF_SIZE )
    {
        printk("[NVREG] ERROR :  the char * key is not terminated with NULL.\n"); 
        return NVREG_ERR_INVALID_PARAM; 
    }

    return nandreg_add( key, val, env.val_len );
}
#endif

int nvreg_get( struct file *file, unsigned long arg )
{
    ST_IOCTL_PARAM	env;
    unsigned long ret;
    int key_len=0;
    int val_len=0;
    char key[NVREG_COPY_BUF_SIZE];
    char val[NVREG_COPY_BUF_SIZE];

    ret = copy_from_user( (void *)&env, (void *)arg, sizeof(ST_IOCTL_PARAM) );
    if( ret < 0 ) printk("[NVREG][get]copy_from_user 1 failed %lu\n", ret ); 

    if( env.key == NULL || env.val == NULL )
    {
        return NVREG_ERR_INVALID_PARAM;
    }

    ret = copy_from_user( key, env.key, NVREG_COPY_BUF_SIZE );
    if( ret < 0 ) printk("[NVREG][get]copy_from_user 2 failed %lu\n", ret ); 

    key_len = strnlen(key, NVREG_COPY_BUF_SIZE);
    if( key_len >= NVREG_COPY_BUF_SIZE )
    {
        printk("[NVREG] ERROR :  the char * key is not terminated with NULL.\n");
        return NVREG_ERR_INVALID_PARAM;
    }
    
    val_len = nandreg_get( key, val, env.val_len );

    if( val_len > 0 )
    {
        if( env.val_len < val_len )
        {
            printk("[NVREG] ERROR : small buffer size(%d) for value length(%d) for %s\n",
                   env.val_len, val_len, key );
            return NVREG_ERR_INVALID_PARAM;
        }
        else
        {
            ret = copy_to_user( env.val, val, val_len );
            if( ret < 0 ) printk("[NVREG][get]copy_to_user 3 failed %lu\n", ret ); 
            return val_len;
        }
    }
    else
    {
        return val_len;
    }

}

#ifdef CONFIG_COMPAT
int nvreg_compat_get( struct file *file, unsigned long arg )
{
    ST_IOCTL_PARAM    env;
    ST_IOCTL_PARAM32  env32;
    unsigned long ret;
    int key_len=0;
    int val_len=0;
    char key[NVREG_COPY_BUF_SIZE];
    char val[NVREG_COPY_BUF_SIZE];

    ret = copy_from_user( (void *)&env32, (void *)arg, sizeof(ST_IOCTL_PARAM32) );
    if( ret < 0 ) printk("[NVREG][get]copy_from_user 1 failed %lu\n", ret );

    env.key = compat_ptr(env32.key);
    env.val = compat_ptr(env32.val);
    env.val_len = env32.val_len;

    if( env.key == NULL || env.val == NULL )
    {
        return NVREG_ERR_INVALID_PARAM;
    }

    ret = copy_from_user( key, env.key, NVREG_COPY_BUF_SIZE );
    if( ret < 0 ) printk("[NVREG][get]copy_from_user 2 failed %lu\n", ret ); 

    key_len = strnlen(key, NVREG_COPY_BUF_SIZE);
    if( key_len >= NVREG_COPY_BUF_SIZE )
    {
        printk("[NVREG] ERROR :  the char * key is not terminated with NULL.\n");
        return NVREG_ERR_INVALID_PARAM;
    }
    
    val_len = nandreg_get( key, val, env.val_len );

    if( val_len > 0 )
    {
        if( env.val_len < val_len )
        {
            printk("[NVREG] ERROR : small buffer size(%d) for value length(%d) for %s\n",
                   env.val_len, val_len, key );
            return NVREG_ERR_INVALID_PARAM;
        }
        else
        {
            ret = copy_to_user( env.val, val, val_len );
            if( ret < 0 ) printk("[NVREG][get]copy_to_user 3 failed %lu\n", ret );
            return val_len;
        }
    }
    else
    {
        return val_len;
    }

}
#endif

int nvreg_del( struct file *file, unsigned long arg )
{
    ST_IOCTL_PARAM	env;
    unsigned long ret;
    char key[NVREG_COPY_BUF_SIZE];

    ret = copy_from_user( (void *)&env, (void *)arg, sizeof(ST_IOCTL_PARAM) );
    if( ret < 0 ) printk("[NVREG][del]copy_from_user 1 failed %lu\n", ret ); 

    if( env.key == NULL )
    {
        return NVREG_ERR_INVALID_PARAM;
    }

    ret = copy_from_user( key, env.key, NVREG_COPY_BUF_SIZE );
    if( ret < 0 ) printk("[NVREG][del]copy_from_user 2 failed %lu\n", ret ); 

    return nandreg_del( key );
}

#ifdef CONFIG_COMPAT
int nvreg_compat_del( struct file *file, unsigned long arg )
{
    ST_IOCTL_PARAM	env;
    ST_IOCTL_PARAM32	env32;
    unsigned long ret;
    char key[NVREG_COPY_BUF_SIZE];

    ret = copy_from_user( (void *)&env32, (void*)arg, sizeof(ST_IOCTL_PARAM32) );
    if( ret < 0 ) printk("[NVREG][del]copy_from_user 1 failed %lu\n", ret );

    env.key = compat_ptr(env32.key);
    env.val = compat_ptr(env32.val);
    env.val_len = env32.val_len;

    if( env.key == NULL )
    {
        return NVREG_ERR_INVALID_PARAM;
    }

    ret = copy_from_user( key, env.key, NVREG_COPY_BUF_SIZE );
    if( ret < 0 ) printk("[NVREG][del]copy_from_user 2 failed %lu\n", ret );

    return nandreg_del( key );
}
#endif

int nvreg_dump( struct file *file, unsigned long arg )
{
    return nandreg_dump();
}

static long nvreg_ioctl( struct file *file, unsigned int command, unsigned long arg )
{
    int ret = NVREG_ERR_INTERNAL;
	
    switch(command)
    {
    case NVREG_IOCTL__ADD:  { ret = nvreg_add(  file, arg); break; }
    case NVREG_IOCTL__GET:  { ret = nvreg_get(  file, arg); break; }
    case NVREG_IOCTL__DEL:  { ret = nvreg_del(  file, arg); break; }
    case NVREG_IOCTL__DUMP: { ret = nvreg_dump( file, arg); break; }

    default:
        printk(" not supported command:%d\n", command);
        break;
	}

    if( ret < 0 )
    {
        printk( "%s(%d) is failed. %s(%d)\n", GetIoctlCmdStr(command), command, GetRetStr(ret), ret );
    }

    return ret;
}

#ifdef CONFIG_COMPAT
static long nvreg_compat_ioctl( struct file *file, unsigned int command, unsigned long arg )
{
    int ret = NVREG_ERR_INTERNAL;
 
    switch(command)
    {
    case NVREG_IOCTL__ADD:  { ret = nvreg_compat_add(  file, arg); break; }
    case NVREG_IOCTL__GET:  { ret = nvreg_compat_get(  file, arg); break; }
    case NVREG_IOCTL__DEL:  { ret = nvreg_compat_del(  file, arg); break; }
    case NVREG_IOCTL__DUMP: { ret = nvreg_dump( file, arg); break; }

    default:
        printk(" not supported command:%d\n", command);
        break;
	}

    if( ret < 0 )
    {
        printk( "%s(%d) is failed. %s(%d)\n", GetIoctlCmdStr(command), command, GetRetStr(ret), ret );
    }

    return ret;
}
#endif

static struct file_operations s_nvreg_drv_fops = {
	.llseek            = NULL,
	.read              = NULL,
	.write             = NULL,
	.poll              = NULL,
	.unlocked_ioctl    = nvreg_ioctl,
#ifdef CONFIG_COMPAT
	.compat_ioctl      = nvreg_compat_ioctl,
#endif
	.mmap              = NULL,
	.open              = NULL,
	.flush             = NULL,
	.release           = NULL,
	.fsync             = NULL,
	.aio_fsync         = NULL,
	.fasync            = NULL,
	.lock              = NULL,
	.sendpage          = NULL,
	.get_unmapped_area = NULL,
	.check_flags       = NULL,
	.flock             = NULL,
};

struct cdev *cdev_info = NULL;
static const int dev_cnt = 1;

static int nvreg_module_init(void)
{
    int rtn;

    nandreg_init();
    nandreg_load();

    /* DeviceNo̎擾 */
    rtn = register_chrdev_region( DEVICE_NUM_NVREG, dev_cnt, DEVICE_NAME_NVREG );
    if( rtn )
    {
        printk("Cannot access to %s \n", DEVICE_NAME_NVREG );
        return NVREG_ERR_DEV_FILE;
    }

    /* cdev̓o^ ---------------------------*/
    cdev_info= cdev_alloc();

    if( cdev_info == NULL )
    {
        printk("%s cannot insmod!\n", DEVICE_NAME_NVREG );
        goto chardev_create_err;
    }

    cdev_info->ops = &s_nvreg_drv_fops;
    cdev_info->owner = THIS_MODULE;
    rtn = cdev_add( cdev_info, DEVICE_NUM_NVREG , dev_cnt );
    if( rtn ){
        printk("%s cannot add device\n", DEVICE_NAME_NVREG );
        goto chardev_create_err;
    }

    return 0;

 chardev_create_err:
    unregister_chrdev_region( DEVICE_NUM_NVREG, dev_cnt );
    return NVREG_ERR_DEV_FILE;

}

static void nvreg_module_exit(void)
{
    cdev_del( cdev_info );
    /* DeviceNỏ */
    unregister_chrdev_region( DEVICE_NUM_NVREG, dev_cnt );
}

module_init(nvreg_module_init);
module_exit(nvreg_module_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Sony Corporation");
MODULE_DESCRIPTION("NON VOLATILE REGISTER");
